from unittest import TestCase, mock
from k9.core import *
from k9.rbac import *
from k9.apps import *

import io
import base64
import os.path
import pprint

import time

from datetime import datetime, timedelta

from k9.pretty_object import PrettyObject

po = PrettyObject()
pp = pprint.PrettyPrinter(indent=2, width=120)

class TestCore(TestCase):

    ###########################################################################
    # Util
    ###########################################################################

    def test_run_command_success(self):
        set_run_output(True)
        result = run_command('ls', '-la')
        self.assertTrue('charts' in result.stdout)

    def test_run_command_fail(self):
        with self.assertRaisesRegex(FileNotFoundError, 'No such file or directory'):
            run_command('bogus -la')

    def test_last_word(self):
        self.assertEqual('my-pod', last_word('pods/my-pod'))

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_view_yaml(self, mock_stdout):
        view_yaml(abs_path('../test/tomcat-deploy-dev.yml'))
        self.assertTrue('tomcat-dev' in mock_stdout.getvalue())

    def test_read_yaml(self):
        body = read_yaml(abs_path('../test/tomcat-deploy-dev.yml'))
        self.assertEqual('tomcat-dev', body['metadata']['name'])

    def test_age(self):
        now = datetime.now(timezone.utc)

        diff = timedelta(hours=14, minutes=4, seconds=22)
        then = now - diff
        self.assertEqual('14:04:22', get_age(then))

        diff = timedelta(hours=4, minutes=14, seconds=2)
        then = now - diff
        self.assertEqual('04:14:02', get_age(then))

        diff = timedelta(hours=23, minutes=14, seconds=2)
        then = now - diff
        self.assertEqual('23:14:02', get_age(then))

        diff = timedelta(days=1, hours=23, minutes=14, seconds=2)
        then = now - diff
        self.assertEqual('1d', get_age(then))

        diff = timedelta(days=25, hours=23, minutes=14, seconds=2)
        then = now - diff
        self.assertEqual('25d', get_age(then))

    def test_absolute_dir(self):
        result = abs_path('test')

        self.assertTrue(len(result) > 4)
        self.assertTrue('/' in result)
        self.assertEqual('test', last_word(result))


    ###########################################################################
    # Namespace
    ###########################################################################

    def test_list_namespaces(self):
        result = list_namespaces()
        self.assertTrue(len(result)>0)

    def test_default_namespace(self):
        set_default_namespace("test")
        self.assertEqual("test", get_default_namespace())

    def test_create_namespace(self):
        try:
            ns = "namespace-unit-test"
            set_default_namespace(ns)

            create_namespace()
            self.assertTrue(namespace_exists())

            result = get_namespace()
            self.assertEqual(ns, result.metadata.name)

        finally:
            delete_namespace()
            self.assertFalse(namespace_exists())

    def test_delete_bogus_namespace(self):
        self.assertEqual(None, delete_namespace('bogus'))

    def test_get_bogus_namespace(self):
        with self.assertRaisesRegex(Exception, 'You must call set_default_namespace()'):
            set_default_namespace('bogus')
            set_default_namespace()
            get_default_namespace()

    ###########################################################################
    # Pods
    ###########################################################################

    def test_list_pods(self):
        try:
            result = list_pods("kube-system")
            self.assertIsNotNone(result)
            self.assertTrue(len(result) > 0)

            create_namespace('asdfjkl')
            set_default_namespace("asdfjkl")
            result = list_pods()
            self.assertTrue(len(result) == 0)
        finally:
            delete_namespace('asdfjkl')


    def test_wait_for_pod(self):
        try:
            set_default_namespace("wait-test")
            if not namespace_exists("wait-test"):
                create_namespace("wait-test")

            if not deployment_exists("tomcat"):
                secret_name = "tomcat-dev"
                secrets = {
                    'ds-url': 'https://some/url',
                    'password': 'My1SecretPassword',
                    'username': 'postgres'
                }
                create_secret(secret_name, secrets)
                body = read_yaml(abs_path('../test/tomcat-deploy-dev.yml'))
                create_deployment(body)
                time.sleep(1)

            pods = list_pods()
            print(pods)
            self.assertEqual(pods[0]['status'], 'Pending')
            wait_for_pod(pods[0]['name'])
            pods = list_pods()
            self.assertEqual(pods[0]['status'], 'Running')
        finally:
            delete_deployment('tomcat-dev')
            delete_secret('tomcat-dev')
            delete_namespace('wait-test')

    def test_copy_to_from_pod(self):
        try:
            # uses example-file.txt
            set_default_namespace("copy-test")
            if not namespace_exists("copy-test"):
                create_namespace("copy-test")
            if not deployment_exists("tomcat"):
                secret_name = "tomcat-dev"
                secrets = {
                    'ds-url': 'https://some/url',
                    'password': 'My1SecretPassword',
                    'username': 'postgres'
                }
                create_secret(secret_name, secrets)
                body = read_yaml(abs_path('../test/tomcat-deploy-dev.yml'))
                create_deployment(body)
                time.sleep(1)
            pods = list_pods()
            print(f'pods are {pods}')
            self.assertIsNotNone(pods)
            wait_for_pod(pods[0]['name'])
            copy_to_pod(pods[0]['name'], abs_path('../test/example-file.txt'), '/var/example-file.txt')
            copy_from_pod(pods[0]['name'], '/var/example-file.txt', abs_path('../test/newex.txt'))
            self.assertTrue(os.path.exists(abs_path('../test/newex.txt')))

        finally:
            delete_deployment('tomcat-dev')
            delete_secret('tomcat-dev')
            delete_namespace('copy-test')
            run_command('rm', '-f', abs_path('../test/newex.txt'))


    ###########################################################################
    # Secrets
    ###########################################################################

    def test_create_secret(self):
        try:
            set_default_namespace("default")

            secret_name = "tomcat-dev"
            secrets = {
                'ds-url': 'https://some/url',
                'password': 'My1SecretPassword',
                'username': 'postgres'
            }

            # Test create_secret()
            result = create_secret(secret_name, secrets)
            self.assertEqual(secret_name, result.metadata.name)

            # Test get_secret()
            result = get_secret(secret_name)
            self.assertEqual(secret_name, result.metadata.name)

            # Check secret values
            for key, value in result.data.items():
                self.assertEqual(secrets[key], base64.b64decode(value).decode('utf8'))

            # Test secret_exists()
            self.assertTrue(secret_exists(secret_name))

            # Test list_secret()
            result = list_secrets()

            secret_list = [
                s['name']
                for s in result
                if s['name'] == secret_name
            ]

            self.assertEqual(1, len(secret_list))
            self.assertEqual(secret_name, secret_list[0])

        finally:
            delete_secret(secret_name)

    def test_secret_exists(self):
        set_default_namespace('default')
        self.assertFalse(secret_exists('bogus-secret'))

    def test_delete_bogus_secret(self):
        set_default_namespace('default')
        self.assertEqual(None, delete_secret('bogus'))

    ###########################################################################
    # Deployments
    ###########################################################################

    def test_deployments(self):
        try:
            secret_name = "tomcat-dev"
            deploy_name = 'tomcat-dev'

            set_default_namespace("deployment-unit2-test")
            if not namespace_exists("deployment-unit2-test"):
                create_namespace()

            secrets = {
                'ds-url': 'https://some/url',
                'password': 'My1SecretPassword',
                'username': 'postgres'
            }

            # Test create_secret()
            create_secret(secret_name, secrets)

            body = read_yaml('tomcat-deploy-dev.yml')
            create_deployment(body)
            self.assertTrue(deployment_exists(deploy_name))

            result = get_deployment(deploy_name)
            self.assertEqual(deploy_name, result.metadata.name)

            result = [
                d
                for d in list_deployments()
                if d['name'] == deploy_name
            ]
            self.assertEqual(1, len(result))
            self.assertFalse(deployment_exists('bogus'))

            # Update deployment
            update_deployment_image(deploy_name, 'tomcat', 'tomcat:8')

            # Confirm that deployment image has been updated.
            result = get_deployment(deploy_name)
            self.assertEqual(deploy_name, result.metadata.name)
            found = [
                container.image
                for container in result.spec.template.spec.containers
                if container.name == 'tomcat'
            ]
            self.assertEqual('tomcat:8', found[0])


            # Scale deployment
            spec = {
                'replicas': 3
            }
            scale_deployment(deploy_name, spec)
            result = get_deployment(deploy_name)
            self.assertEqual(deploy_name, result.metadata.name)
            self.assertEqual(3, result.spec.replicas)


        finally:

            delete_deployment(deploy_name)
            delete_secret(secret_name)
            delete_namespace()


    def test_delete_bogus_deployment(self):
        set_default_namespace('default')
        self.assertEqual(None, delete_deployment('bogus'))

    ###########################################################################
    # Services
    ###########################################################################
    def test_service(self):
        try:
            # Arrange
            svc_name = 'tomcat-svc-dev'
            secret_name = "tomcat-dev"
            deploy_name = 'tomcat-dev'

            set_default_namespace("service-unit-test")
            if not namespace_exists("service-unit-test"):
                create_namespace()

            secrets = {
                'ds-url': 'https://some/url',
                'password': 'My1SecretPassword',
                'username': 'postgres'
            }

            create_secret(secret_name, secrets)
            body = read_yaml(abs_path('../test/tomcat-deploy-dev.yml'))
            create_deployment(body)

            # Act
            body = read_yaml(abs_path('../test/tomcat-svc-dev.yml'))

            create_service(body)

            # Assert
            result = get_service(svc_name)
            self.assertEqual(svc_name, result.metadata.name)

            result = list_services()
            found = [
                svc['name']
                for svc in result
                if svc_name in svc['name']
            ]
            self.assertEqual(1, len(found))

        finally:
            if service_exists(svc_name):
                delete_service(svc_name)

            delete_deployment(deploy_name)

            delete_secret(secret_name)
            delete_namespace()

    def test_service_exists_fail(self):
        set_default_namespace('default')
        self.assertFalse(service_exists('bogus'))

    def test_delete_bogus_service(self):
        set_default_namespace('default')
        self.assertEqual(None, delete_service('bogus'))

    ###########################################################################
    # Service Accounts
    ###########################################################################

    def test_service_accounts(self):
        try:
            ############
            # Arrange
            set_default_namespace('default')
            sa_name = "unit-test-tomcat-sa"

            ############
            # Act
            result = create_service_account(sa_name)

            ############
            # Assert

            self.assertEqual(sa_name, result.metadata.name)

            # test get_service_account()
            result = get_service_account(sa_name)
            self.assertEqual(sa_name, result.metadata.name)

            # test service_account_exists()
            self.assertTrue(service_account_exists(sa_name))

            # test_list_service_accounts()
            result = list_service_accounts()

            result = [
                sa['name']
                for sa in result
                if sa['name'] == sa_name
            ]
            self.assertEqual(1, len(result))

        finally:
            delete_service_account(sa_name)


    def test_service_account_exists_fail(self):
        set_default_namespace('default')
        self.assertFalse(service_account_exists('bogus'))
        self.assertTrue(service_account_exists('default'))

    def test_list_service_account_fail(self):
        self.assertEqual(0, len(list_service_accounts('bogus')))

    def test_delete_bogus_service_account(self):
        set_default_namespace('default')
        self.assertEqual(None, delete_service_account('bogus'))




from unittest import TestCase
from k9.storage import *
import subprocess

class TestStorage(TestCase):
    def test_create_storage_class(self):
        name = 'aws-storage-class'

        print("")
        create_storage_class('storage.yml')

        self.assertTrue(storage_class_exists(name))

        delete_storage_class(name)

        self.assertFalse(storage_class_exists(name))

    def test_create_storage_class_fail(self):
        with self.assertRaisesRegex(subprocess.CalledProcessError, 'returned non-zero exit status 1.'):
            create_storage_class('bogus')

    def test_delete_storage_class_fail(self):
        with self.assertRaisesRegex(subprocess.CalledProcessError, 'returned non-zero exit status 1.'):
            delete_storage_class('bogus')




class PrettyObject:
    """
    Prints in an easily readable, simplified, JSON format.  It allows customization of how big dictionaries,
    lists, sets and tuples can be before they are dropped to the next line.

    Note that this is a very relaxed JSON format, and is syntactly incorrect.  You'll want to 
    use **pprint** if you need strict JSON formatted output.
    """

    # Establishes size that list structures would extend
    # to line by line display

    dictLimit  = 1   #: max number of dictionary entries on a single line, default is 1
    listLimit  = 1   #: max number of list items on a single line, default is 1
    setLimit   = 1   #: max number of items in a set on a single line, default is 1
    tupleLimit = 1   #: max number of items in a tuple on a single line, default is 1

    INDENT = "    "

    def pretty_dict(self, obj: dict, indent=""):
        if len(obj) < self.dictLimit:
            return self.pretty_dict_line(obj, indent)

        result = "{\n"
        for key in obj:
            result += indent + self.INDENT + key + ": ";
            result += self.string(obj[key], indent + self.INDENT) + "\n"

        result += indent + "}"
        return result

    def pretty_dict_line(self, obj: dict, indent=""):
        if len(obj) == 0:
            return "{}"

        first = True
        result = "{"
        for key in obj:
            if not first:
                result += ", "
            first = False
            result += key + ": ";
            result += self.string(obj[key], indent)

        result += "}"
        return result

    def pretty_list(self, obj, indent=""):
        if len(obj) < self.listLimit:
            return self.pretty_list_line(obj, indent)

        result = "[\n"
        for item in obj:
            result += indent + self.INDENT + self.string(item, indent + self.INDENT) + "\n"

        result += indent + "]"
        return result

    def pretty_list_line(self, obj, indent=""):
        if len(obj) == 0:
            return "[]"

        first = True;
        result = "["
        for item in obj:
            if not first:
                result += ", "
            first = False
            result += self.string(item, indent)

        result += "]"

        return result

    def pretty_tuple(self, obj, indent=""):
        if len(obj) < self.listLimit:
            return self.pretty_tuple_line(obj, indent)

        result = "(\n"
        for item in obj:
            result += indent + self.INDENT + self.string(item, indent + self.INDENT) + "\n"

        result += indent + ")"
        return result

    def pretty_tuple_line(self, obj, indent=""):
        if len(obj) == 0:
            return "()"

        first = True;
        result = "("
        for item in obj:
            if not first:
                result += ", "
            first = False
            result += self.string(item, indent)

        result += ")"

        return result

    def string(self, obj, indent=""):
        if isinstance(obj, dict):
            return self.pretty_dict(obj, indent)

        if isinstance(obj, list):
            return self.pretty_list(obj, indent)

        if isinstance(obj, tuple):
            return self.pretty_tuple(obj, indent)

        if isinstance(obj, str):
            return '"' + obj + '"'

        else:
            return str(obj)

    def print(self, obj):
        """
        Prints an object in a simplified JSON format for easy viewing.  Note that if you need
        syntactically correct JSON, you should use **pprint** instead.

        :param obj: The object to print, can be either list or dictionary.

        Example::

            po = PrettyObject()

            obj = [
                {"Name": "default-vpc", "VpcId": "vpc-2eb67c54", "CidrBlock": "172.31.0.0/16", "IsDefault": True, "State": "available"},
                {"Name": "dvl-vpc", "VpcId": "vpc-422314113412", "CidrBlock": "10.0.0.0/16", "IsDefault": False, "State": "pending"}
            ]
            po.print(obj)

        Output::

            [
                {
                    Name: "default-vpc"
                    VpcId: "vpc-2eb67c54"
                    CidrBlock: "172.31.0.0/16"
                    IsDefault: True
                    State: "available"
                }
                {
                    Name: "dvl-vpc"
                    VpcId: "vpc-422314113412"
                    CidrBlock: "10.0.0.0/16"
                    IsDefault: False
                    State: "pending"
                }
            ]

        """
        print(self.string(obj))

    def print_table(self, obj):
        """
        Prints a list of dictionary objects in a table format.

        :param obj: A list of dictionary objects.

        Example::

            po = PrettyObject()

            obj = [
                {"Name": "default-vpc", "VpcId": "vpc-2eb67c54", "CidrBlock": "172.31.0.0/16", "IsDefault": True, "State": "available"},
                {"Name": "dvl-vpc", "VpcId": "vpc-422314113412", "CidrBlock": "10.0.0.0/16", "IsDefault": False, "State": "pending"}
            ]
            po.print_table(obj)

        Output::

            Name         VpcId             CidrBlock      IsDefault  State
            -----------  ----------------  -------------  ---------  ---------
            default-vpc  vpc-2eb67c54      172.31.0.0/16  True       available
            dvl-vpc      vpc-422314113412  10.0.0.0/16    False      pending
        """
        if not isinstance(obj, list):
             print("Object is not a table:")
             self.print(obj)
             return

        # Capture stats on the table
        lengths = {}
        order = []
        for record in obj:
            for key in record:
                value  = str(record[key])
                length = len(value)

                if not key in lengths:
                    keylen = len(key)
                    if keylen > length:
                        length = keylen

                    lengths.update({key: length})
                    order.append(key)
                    continue

                if lengths[key] < length:
                    lengths.update({key: length})

        # Print the header
        line = ""
        for key in order:
            line += key.ljust(lengths[key])
            line += "  "
        print(line)

         # Print underline
        line = ""
        for key in order:
            line += "".ljust(lengths[key], '-')
            line += "  "
        print(line)

        # Print table content
        for record in obj:
            line = ""
            for key in order:
                if key in record:
                    value = str(record[key])
                else:
                    value = ""
                line += value.ljust(lengths[key]+2)

            print(line)


def get_tags(tag):
    result = {}
    for vpair in tag:
        result.update({vpair.get('Key'): vpair.get('Value')})

    return result


def get_tag(tags, key):
    for vpair in tags:
        if vpair['Key'] == key:
            return vpair['Value']
    return ""

def get_name(tags):
    if tags is None:
        return ''
    for vpair in tags:
        if vpair['Key'] == 'Name':
            return vpair['Value']
    return ''

def get_true(value):
    if value:
        return 'True'
    else:
        return ''




from unittest import TestCase, mock
from k9.pretty_object import PrettyObject, get_tag, get_tags, get_name, get_true
import io

class TestPrettyObject(TestCase):

    def test_get_tags(self):
        tags = [{"Key": "Name", "Value": "default-vpc"}, {"Key": "Author", "Value": "simon.woo"}]
        value = get_tags(tags)
        self.assertEqual("default-vpc", value["Name"] )
        self.assertEqual("simon.woo", value["Author"] )

    def test_get_tag(self):
        tags = [{"Key": "Name", "Value": "default-vpc"}]
        value = get_tag(tags, "Name")
        self.assertEqual( "default-vpc", value )

    def test_get_name(self):
        tags = [{"Key": "Name", "Value": "default-vpc"}]
        value = get_name(tags)
        self.assertEqual("default-vpc", value)
        value = get_name([])
        self.assertEqual('', value)

    def test_get_true(self):
        value = get_true(True)
        self.assertEqual('True', value)
        value = get_true(False)
        self.assertEqual('', value)

    def test_pretty_dictionary(self):
        po = PrettyObject()
        obj = {"test": "test", "test2": "test2"}
        result = po.string(obj)
        self.assertEqual('{\n'
                         '    test: "test"\n'
                         '    test2: "test2"\n'
                         '}', result)

        obj = {"test": [], "test2": "test2"}
        result = po.string(obj)
        self.assertEqual('{\n'
                         '    test: []\n'
                         '    test2: "test2"\n'
                         '}', result)

        obj = {"test": {"test": "test"}}
        result = po.string(obj)
        self.assertEqual('{\n'
                         '    test: {\n'
                         '        test: "test"\n'
                         '    }\n'
                         '}', result)

        obj = {"test": {"test": "test", "test2": {"test3": "test3"}}}
        result = po.string(obj)
        self.assertEqual('{\n'
                         '    test: {\n'
                         '        test: "test"\n'
                         '        test2: {\n'
                         '            test3: "test3"\n'
                         '        }\n'
                         '    }\n'
                         '}', result)

    def test_pretty_dictionary_line(self):
        po = PrettyObject()
        po.dictLimit = 10
        obj = {"test": "test", "test2": "test2"}
        result = po.string(obj)
        self.assertEqual('{test: "test", test2: "test2"}', result)

        obj = {"test": {"test": "test"}}
        result = po.string(obj)
        self.assertEqual('{test: {test: "test"}}', result)

        obj = {"test": {"test": "test", "test2": {"test3": "test3"}}}
        result = po.string(obj)
        self.assertEqual('{test: {test: "test", test2: {test3: "test3"}}}', result)

    def test_pretty_list(self):
        po = PrettyObject()

        obj = ["test1", "test2"]
        result = po.string(obj)
        self.assertEqual('[\n'
                         '    "test1"\n'
                         '    "test2"\n'
                         ']', result)

        obj = ["test1", {"test": "test2"}]
        result = po.string(obj)
        self.assertEqual('[\n'
                         '    "test1"\n'
                         '    {\n'
                         '        test: "test2"\n'
                         '    }\n'
                         ']', result)

    def test_pretty_list_line(self):
        po = PrettyObject()
        po.dictLimit = 10
        po.listLimit = 10
        obj = ["test1", "test2"]
        result = po.string(obj)
        self.assertEqual('["test1", "test2"]', result)

        obj = ["test1", {"test": "test"}]
        result = po.string(obj)
        self.assertEqual('["test1", {test: "test"}]', result)

    def test_pretty_tuple(self):
        po = PrettyObject()

        obj = ("test1", "test2")
        result = po.string(obj)
        self.assertEqual('(\n'
                         '    "test1"\n'
                         '    "test2"\n'
                         ')', result)

        obj = ("test1", {"test": "test2"})
        result = po.string(obj)
        self.assertEqual('(\n'
                         '    "test1"\n'
                         '    {\n'
                         '        test: "test2"\n'
                         '    }\n'
                         ')', result)

    def test_pretty_tuple_line(self):
        po = PrettyObject()
        po.dictLimit = 10
        po.listLimit = 10
        obj = ("test1", "test2")
        result = po.string(obj)
        self.assertEqual('("test1", "test2")', result)

        obj = ("test1", {"test": "test"})
        result = po.string(obj)
        self.assertEqual('("test1", {test: "test"})', result)

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_print(self, mock_stdout):
        po = PrettyObject()

        obj = [
            {"Name": "default-vpc", "VpcId": "vpc-2eb67c54", "CidrBlock": "172.31.0.0/16", "IsDefault": True, "State": "available"},
            {"Name": "dvl-vpc", "VpcId": "vpc-422314113412", "CidrBlock": "10.0.0.0/16", "IsDefault": False, "State": "pending"}
        ]
        po.print(obj)
        self.assertTrue('vpc-2eb67c54' in mock_stdout.getvalue())
        self.assertTrue('vpc-422314113412' in mock_stdout.getvalue())

    @mock.patch('sys.stdout', new_callable=io.StringIO)
    def test_print_table(self, mock_stdout):
        po = PrettyObject()

        obj = [
            {"Name": "default-vpc", "VpcId": "vpc-2eb67c54", "CidrBlock": "172.31.0.0/16", "IsDefault": True, "State": "available"},
            {"Name": "dvl-vpc", "VpcId": "vpc-422314113412", "CidrBlock": "10.0.0.0/16", "IsDefault": False, "State": "pending"}
        ]
        po.print_table(obj)
        self.assertTrue('vpc-2eb67c54' in mock_stdout.getvalue())
        self.assertTrue('vpc-422314113412' in mock_stdout.getvalue())



from collections import namedtuple

from kubernetes import client, config
from k9.pretty_object import PrettyObject

from datetime import datetime, timezone
import os
import pprint
import subprocess
import time
import yaml

po = PrettyObject()
if os.getenv('KUBERNETES_SERVICE_HOST'):
  config.load_incluster_config()
else:
  config.load_kube_config()

v1_core = client.CoreV1Api()

default_namespace = None


###########################################################################
# Utility Functions
###########################################################################
CommandResult = namedtuple('CommandResult', 'args returncode stdout stderr')
output_on = True


def set_run_output(output_mode=True):
    """
    Sets flag to indicate whether outputs from the run_command should be printed
    to the console or not.
    :param output_mode: True if stdout should be printed to console.  Default is True
    """
    global output_on
    output_on = output_mode


def run_command(*params):
    """
    This is a helper function to make it easier to run shell commands in
    a consistent manner.

    Calls the subprocess.run() command with the following options:

    capture_output=True  Which provides **stdout** and **stderr**
    check=True           Which throws an exception if an error occurs.  The exception
                         type `subprocess.CalledProcessError <https://docs.python.org/3/library/subprocess.html#subprocess.SubprocessError>`_

    :param params: a variable list of parameters
    :return: A named tuple version of `subprocess.CompletedProcess <https://docs.python.org/3/library/subprocess.html>`_
             that translates the stdout and stderr with str( 'utf-8')

    Example Call::
        result = run_command('ls', '-la')
        print(f'result.args: {result.args}')
        print(f'result.returncode: {result.returncode}')
        print(f'result.stdout: {result.stdout}')
        print(f'result.stderr: {result.stderr}')


    """
    result = subprocess.run(params, capture_output=True, check=True)

    stdout = ""
    if result.stdout:
        stdout = str(result.stdout, 'utf-8')
        if output_on:
            print(stdout)

    stderr = ""
    if result.stderr:
        stderr = str(result.stderr, 'utf-8')

    return CommandResult(
        args=result.args,
        returncode=result.returncode,
        stdout=stdout,
        stderr=stderr
    )

def last_word(value: str):
    """
    Splits out the word after the last slash in a string.  K8
    objects are expressed in a path style name

    Example
    -------

    >>> last_word('pods/my-pod')
    'my-pod'
    """
    return value.split('/')[-1:][0]


def view_yaml(fn: str):
    """
    Dumps out yaml file in JSON format for easy viewing. This
    is useful when constructing the body of a request that matches a known yaml format.

    Example
    -------

    >>> view_yaml('tomcat-deploy-dev.yml')
    """
    file = None
    try:
        file = open(fn)
        pprint.pprint(yaml.safe_load(file))

    finally:
        if file is not None:
            file.close()


def read_yaml(fn: str):
    """
    Reads a YAML file and returns the resulting the object.

    Example
    -------

    >>> read_yaml('tomcat-deploy-dev.yml')

    """
    file = None
    try:
        file = open(fn)
        return yaml.safe_load(file)

    finally:
        if file is not None:
            file.close()


def get_age(creation_time: datetime):
    """
    Given a creation timestamp, return the difference in time from now.

    :param creation_time: The time we want to measure age from
    :return: timedelta - the amount of time since creation_time
    """
    now = datetime.now(timezone.utc)
    delta = now - creation_time

    if delta.days > 0:
        return f'{delta.days}d'

    hours = delta.seconds/3600

    seconds = delta.seconds%3600

    minutes = seconds/60
    seconds = seconds % 60

    return '%02d:%02d:%02d' % (hours, minutes, seconds)


def abs_path(file: str):
    """
    Sets an absolute path relative to the **k9** package directory.

    Example::
        result = abs_path('myfile)

    Result::
        /Users/simon/git/k9/k9/myfile


    This is used primarily for building unit tests within the K9 package
    and is not expected to be useful to K9 library users.

    :param file: File or directory to attach absolute path with
    :return: absolute path to specified file or directory
    """
    basedir = os.path.abspath(os.path.dirname(__file__))
    return os.path.join(basedir, file)

###########################################################################
# Namespaces
###########################################################################

def set_default_namespace(namespace: str = None):
    """
    Sets the default namespace for all functions that need namespace.

    Most of the functions in this library will require a namespace parameter.
    If the namespace is not provided, the default namespace you set will
    be used instead, simplifying the call.

    Typically, this should be one of the first calls you make when
    working with this library.

    :param: The name of the default namespace. If none, we use "default"
    """
    global default_namespace
    if namespace is None:
      namespace = 'default'
    default_namespace = namespace


def get_default_namespace():
    """
    Gets the default namespace set using set_default_namespace().

    :return: The default namespace's name
    """
    global default_namespace
    if default_namespace is None:
        raise Exception("You must call set_default_namespace() first before using most of the K9 API functions.")

    return default_namespace


def list_namespaces():
    """
    Retrieves a list of namespaces and associated status.  Returns same
    information as `kubectl get namespaces`

    :return: list of dictionaries with **name** and **status**
    """

    return [
        {
            'name': namespace.metadata.name,
            'status': namespace.status.phase,
            'age': get_age(namespace.metadata.creation_timestamp)
        }
        for namespace in v1_core.list_namespace().items
    ]

def get_namespace(namespace: str = None):
    """
    Gets an object that holds detailed information about the supplied namespace.

    :param namespace: Namespace to retrieve.  If None, will use default namespace.
    :return: `V1Namespace <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Namespace.md>`_
    """
    if namespace is None:
        namespace = get_default_namespace()

    return v1_core.read_namespace(namespace)


def namespace_exists(namespace: str = None):
    """
    Determines if the specified namespace exists.

    :param namespace: The namespace to check for.  If None, then the default namespace is used.
    :return: bool - True if namespace exists
    """
    try:
        if namespace is None:
            namespace = get_default_namespace()

        result = get_namespace(namespace)
        return result.status.phase == 'Active' and result.metadata.name == namespace

    except:
        return False

def create_namespace(namespace: str = None):
    """
    Creates the specified namespace.

    :param namespace: Specifies the namespace to create.  If None, then the default namespace is used.
    :return: `V1Namespace <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Namespace.md>`_
    """

    if namespace is None:
        namespace = get_default_namespace()

    body = \
        {
            "apiVersion": "v1",
            "kind": "Namespace",
            "metadata": {
                "name": namespace,
            }
        }

    return v1_core.create_namespace(body)

def delete_namespace(namespace: str = None):
    """
    Deletes the specified namespace.

    :param namespace: Namespace to delete.  If None, the default namespace is used.
    :return: None if namespace doesn't exist, otherwise `V1Status <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Status.md>`_
    """
    if namespace is None:
        namespace = get_default_namespace()

    if not namespace_exists(namespace):
        return None

    result = v1_core.delete_namespace(namespace)

    # wait for namespace to be gone
    try:
        while get_namespace(namespace) is not None:
           time.sleep(1)

    except client.rest.ApiException as e:
        # spec found here: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/rest.py
        if not "Not Found" in e.reason:
            raise


###########################################################################
# Nodes
###########################################################################

def list_nodes():
    """
    Retrieve a list of nodes
    
    :return: Return a list of nodes - each node represented by a dictionary with **name**, **status**, **ip**, **label**, **reason**, and **start_time**
    """
    return [
        {
            'name': node.metadata.name,
            'status': node.status.phase,
            'info': node.status.node_info
        }
        for node in v1_core.list_node().items
    ]

def get_node(name: str):
    """

    :param name: The name of the node
    :return: Return information related to the node
    """
    selector = 'metadata.name=' + name
    return [
        {
            'name': node.metadata.name,
            'status': node.status.phase,
            'info': node.status.node_info
        }

        for node in v1_core.list_node(field_selector=selector).items
    ]

###########################################################################
# Pods
###########################################################################

def list_pods(namespace: str = None):
    """
    List all pods in a given namespace

    :param namespace: Namespace to search.  If None, uses the default namespace
    :return: Returns a list of pods - each pod represented by a dictionary with **name**, **status**, **ip**, **node**, **labels**, **reason**, and **start_time**
    """
    if namespace is None:
        namespace = get_default_namespace()

    pod_list = v1_core.list_namespaced_pod(namespace)
    ret = []
    for pod in pod_list.items:
        ready = False
        if not pod.status.start_time:
          pod.status.start_time = datetime.now(timezone.utc)
        for cond in pod.status.conditions:
            if cond.type == 'Ready':
                ready = (cond.status == 'True')
                break
        ret.append({
              'name': pod.metadata.name,
              'status': pod.status.phase,
              'ready': ready,
              'ip': pod.status.pod_ip,
              'labels': pod.metadata.labels,
              'reason': pod.status.reason,
              'node': pod.spec.node_name,
              'age': get_age(pod.status.start_time)
        })
    return ret

def copy_to_pod(pod: str, host_fn: str, dest_fn: str, namespace: str = None):
    """
    Copies a file to the specified pod.
    
    Example::
        copy_to_pod('jenkins-12345', '.output/jenkins_config.xml', '/var/jenkins_home/some_config.xml')
    
    :param pod: The name of the pod you're copying the file into
    :param namespace: The namespace the pod is found in (if not present, we search in the default ns)
    :param host_fn: The filename, relative to the execution folder
    :param dest_fn: The desired location + filename on the pod.

    This function doesn't return anything if successful. It throws errors if the file does not exist on the host, the file already exists on the pod, or if the pod name is invalid.
    """
    try:
        if not namespace:
            namespace = get_default_namespace()
        run_command('kubectl', 'cp', host_fn, f'{namespace}/{pod}:{dest_fn}')
    except Exception as e:
        print(e)

def copy_from_pod(pod: str, host_fn: str, dest_fn: str, namespace: str = None):
    """
    This is used for copying a file from inside a pod to the host machine.
    
    Example::
        copy_from_pod('jenkins-12345', '/var/jenkins_home/some_config.xml', '.output/jenkins_config.xml')

    :param pod: The name of the pod you're copying the file from
    :param namespace: The namespace the pod is found in (if not present, we search in the default ns)
    :param host_fn: The location + filename on the pod.
    :param dest_fn: The destination file relative to the folder we're executing from.

    This function doesn't return anything if successful. It throws errors if the pod or file (on the pod) doesn't exist. This overwrites files on the host machine, unlike copy_to_pod()
    """
    try:
        if not namespace:
            namespace = get_default_namespace()
        run_command('kubectl', 'cp', f'{namespace}/{pod}:{host_fn}', dest_fn)
    except Exception as e:
        print(e)

def wait_for_pod(pod: str, namespace: str = None, interval: int = 10, timeout: int = 240):
    """
    Waits for a pod to be ready. Generally used when we need to use an API or get a file to/from the pod for configuration.

    Example::
        wait_for_pod('jenkins-asdf123')
    
    :param pod: The name of the pod we're waiting on
    :param namespace: The namespace the pod is found in (if not present, we search in the default ns)
    :param interval: How long to wait between checks (in seconds)
    :param timeout: How many seconds to wait (in total) before timing out the operation

    :return: If the pod is ready/running (True) or if it isn't ready by the timeout (False)
    """
    time.sleep(5)
    ready = False
    for i in range(timeout // interval + 1):
        pods = list_pods(namespace)
        for p in pods:
            if p['name'] == pod:
                ready = p['ready']
        if ready:
            return True
        if i <= timeout // interval:
            print(f'waiting for {pod} to be ready')
            time.sleep(interval)
    return False

###########################################################################
# Secrets
###########################################################################

def list_secrets(namespace: str = None):
    """
    Lists secrets in a given namespace.

    :param namespace: Namespace you want to search.  If None, the default namespace is used.
    :return: A list of dictionaries with **name**, **type**, **data** (for number of entries), and **age**
    """
    if namespace is None:
        namespace = get_default_namespace()

    return [
        {
            'name': secret.metadata.name,
            'type': secret.type,
            'data': len(secret.data),
            'age': get_age(secret.metadata.creation_timestamp)
        }
        for secret in v1_core.list_namespaced_secret(namespace).items
    ]


def secret_exists(name: str, namespace: str = None):
    """
    Determines if a secret exists.

    :param name: Name of secret
    :param namespace: Namespace to search, if None, uses default namespace
    :return: True if specfied secret exists.
    """
    if namespace is None:
        namespace = get_default_namespace()

    try:
        result = get_secret(name, namespace)
        return result.metadata.name == name

    except:
        return False


def get_secret(name: str, namespace: str = None):
    """
    Gets a secret's metadata and value(s).

    :param name: Name of secret
    :param namespace: Namespace to search, if None, uses default namespace
    :return: None if not found, otherwise `V1Secret <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Secret.md>`_
    """
    if namespace is None:
        namespace = get_default_namespace()

    return v1_core.read_namespaced_secret(name, namespace)


def create_secret(name: str, secrets: dict, namespace: str = None):
    """
    Creates a secret.

    :param name: Name of secret
    :param secrets: Dictionary containing name value pairs of secrets
    :param namespace:
    :return: `V1Secret <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Secret.md>`_

    Example::

        # Note that you should not embed secrets in your source code, this is
        # simply to illustrate how the function call works.
        secret_name = "tomcat-dev"
        secrets = {
            'ds-url': 'https://some/url',
            'password': 'My1SecretPassword',
            'username': 'postgres'
        }

        # Test create_secret()
        result = create_secret(secret_name, secrets)

    Output::

        {'api_version': 'v1',
         'data': {'ds-url': 'aHR0cHM6Ly9zb21lL3VybA==',
                  'password': 'TXkxU2VjcmV0UGFzc3dvcmQ=',
                  'username': 'cG9zdGdyZXM='},
         'kind': 'Secret',
         'metadata': {'annotations': None,
                      'cluster_name': None,
                      'creation_timestamp': datetime.datetime(2019, 10, 17, 17, 20, 56, tzinfo=tzutc()),
                      'deletion_grace_period_seconds': None,
                      'deletion_timestamp': None,
                      'finalizers': None,
                      'generate_name': None,
                      'generation': None,
                      'initializers': None,
                      'labels': None,
                      'managed_fields': None,
                      'name': 'tomcat-dev',
                      'namespace': 'default',
                      'owner_references': None,
                      'resource_version': '2053051',
                      'self_link': '/api/v1/namespaces/default/secrets/tomcat-dev',
                      'uid': '7ab378c0-f102-11e9-a715-025000000001'},
         'string_data': None,
         'type': 'Opaque'}
    """
    if namespace is None:
        namespace = get_default_namespace()

    body = client.V1Secret()
    body.api_version = 'v1'
    body.kind = 'Secret'
    body.metadata = {'name': name}
    body.string_data = secrets
    body.type = 'Opaque'

    return v1_core.create_namespaced_secret(namespace, body)


def delete_secret(name: str, namespace: str = None):
    """
    Delete specified secret.

    :param name: Name of secret to delete.
    :param namespace: Namespace to delete from.  If None, default namespace is used.
    :return: `V1Status <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Status.md>`_ if secret exists, if not, None.
    """
    if namespace is None:
        namespace = get_default_namespace()

    if secret_exists(name, namespace):
        return v1_core.delete_namespaced_secret(name, namespace)
    else:
        return None


###########################################################################
# Services
###########################################################################

def list_services(namespace: str = None):
    """
    Lists the services in the namespace.

    :param namespace: Namespace to list services from.  If None, default namespace will be used.
    :return: List of dictionaries with **name**, **type**, **cluster-ip**, **external-ips**, **ports**, and **age**
    """
    if namespace is None:
        namespace = get_default_namespace()

    return [
        {
            'name': svc.metadata.name,
            'type': svc.spec.type,
            'cluster-ip': svc.spec.cluster_ip,
            'external-ips': svc.spec.external_i_ps,
            'ports': [
                f'{port.target_port}/{port.protocol}'
                for port in svc.spec.ports
            ],
            'age': get_age(svc.metadata.creation_timestamp)
        }
        for svc in v1_core.list_namespaced_service(namespace).items
    ]


def service_exists(name: str, namespace: str = None):
    """
    Checks existence of specified service.

    :param name: Name of service.
    :param namespace: Namespace to get service from.  If None, will use default namespace.
    :return: True if service exists.
    """
    if namespace is None:
        namespace = get_default_namespace()
    try:
        result = get_service(name, namespace)

        return result.metadata.name == name
    except:

        return False;

def get_service(name: str, namespace: str = None):
    """
    Retrieves details on the specified service.

    :param name: Name of service to get.
    :param namespace: Namespace to get service from.  If None, will use default namespace.
    :return: `V1Service <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Service.md>`_
    """
    if namespace is None:
        namespace = get_default_namespace()

    return v1_core.read_namespaced_service(name, namespace)

def create_service(body: dict, namespace: str = None):
    """
    Creates a service based on definition provided by **body**.

    :param body:  Service description: https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Service.md
    :param namespace: Namespace to create service in.  If None, will use default namespace.
    :return: `V1Service <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Service.md>`_

    You'll most likely create a YAML file to describe your service and read that in.  You can use
    read_yaml() to generate the body as follows:

    Example::

        result = create_service(read_yaml('my-service.yaml'))

    Sample YAML file::

        apiVersion: v1
        kind: Service
        metadata:
          name: tomcat-svc-dev
          labels:
                svc: tomcat
                env: dev
        spec:
          type: ClusterIP
          ports:
          - port: 8080
            targetPort: 8080
            protocol: TCP
          selector:
                app: tomcat
                env: dev
    """
    if namespace is None:
        namespace = get_default_namespace()

    return v1_core.create_namespaced_service(namespace=namespace, body=body)


def delete_service(name: str, namespace: str = None):
    """
    Deletes the specified service.  This function will check whether the service exists before attempting the delete.

    :param name: Name of service to delete
    :param namespace: Namespace to delete from.  If None, default namespace is used.
    :return: None if service doesn't exist, otherwise `V1Status <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Status.md>`_
    """
    if namespace is None:
        namespace = get_default_namespace()

    if service_exists(name, namespace):
        return v1_core.delete_namespaced_service(name, namespace)
    else:
        return None



###########################################################################
# Service Accounts
###########################################################################

def list_service_accounts(namespace: str = None):
    """
    Lists service accounts in a namespace.

    :param namespace: The namespace to list accounts from. If None, we use the default namespace.
    :return: A list of every service account in the namespace.

    Example Output::
    
        [
          {
            'name': 'elastic-sa',
            'age': 123456789
          },
          {
            'name': 'fluent-sa',
            'age': 123456999
          }
        ]
    """
    if namespace is None:
        namespace = get_default_namespace()

    result = v1_core.list_namespaced_service_account(namespace)

    return [
        {
            'name': sa.metadata.name,
            'age': get_age(sa.metadata.creation_timestamp)
        }
        for sa in result.items
    ]


def create_service_account(name: str, namespace: str = None):
    """
    Create a service account.

    :param name: Name of service account.
    :param namespace: namespace to create service account in.  If None, create service account in default namespace
    :return: `V1ServiceAccount <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1ServiceAccount.md>`_
    """
    if namespace is None:
        namespace = get_default_namespace()

    body =\
        { 'apiVersion': 'v1',
          'kind': 'ServiceAccount',
          'metadata' : { 'name': name }
        }

    return v1_core.create_namespaced_service_account(namespace, body)

def get_service_account(name: str, namespace: str = None):
    """
    Get details of specified service account

    :param name: Name of service account to retrieve.
    :param namespace: Namespace to retrieve service account from. If None, retrieve from default namespace.
    :return: `V1ServiceAccount <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1ServiceAccount.md>`_
    """
    if namespace is None:
        namespace = get_default_namespace()
    pass

    return v1_core.read_namespaced_service_account(name, namespace)


def service_account_exists(name: str, namespace: str = None):
    """
    Checks for existence of service account

    :param name: Name of service account to look for.
    :param namespace: Namespace to look in.  If None, look in default namespace.
    :return: True if specified account exists.
    """
    if namespace is None:
        namespace = get_default_namespace()

    try:
        result = get_service_account(name, namespace)
        return result.metadata.name == name

    except:
        return False


def delete_service_account(name: str, namespace: str = None):
    """
    Delete specified service account.

    :param name: Name of service account to delete.
    :param namespace: Namespace to delete from.  If None, delete from default namespace.
    :return: None if service account doesn't exists, otherwise returns `V1Status <https://github.com/kubernetes-client/python/blob/master/kubernetes/docs/V1Status.md>`_
    """
    if namespace is None:
        namespace = get_default_namespace()

    if service_account_exists(name, namespace):
        return v1_core.delete_namespaced_service_account(name, namespace)
    else:
        return None



from unittest import TestCase, mock
from k9.apps import *

class TestApps(TestCase):

    ###########################################################################
    # Deployments
    ###########################################################################

    def test_deployments(self):
        try:
            secret_name = "tomcat-dev"
            deploy_name = 'tomcat-dev'

            set_default_namespace("deployment-unit-test")
            if not namespace_exists("deployment-unit-test"):
                create_namespace()

            secrets = {
                'ds-url': 'https://some/url',
                'password': 'My1SecretPassword',
                'username': 'postgres'
            }

            # Test create_secret()
            create_secret(secret_name, secrets)

            body = read_yaml(abs_path('../test/tomcat-deploy-dev.yml'))
            create_deployment(body)
            self.assertTrue(deployment_exists(deploy_name))

            result = get_deployment(deploy_name)
            self.assertEqual(deploy_name, result.metadata.name)

            result = [
                d
                for d in list_deployments()
                if d['name'] == deploy_name
            ]
            self.assertEqual(1, len(result))
            self.assertFalse(deployment_exists('bogus'))

            # Update deployment
            update_deployment_image(deploy_name, 'tomcat', 'tomcat:8')

            # Confirm that deployment image has been updated.
            result = get_deployment(deploy_name)
            self.assertEqual(deploy_name, result.metadata.name)
            found = [
                container.image
                for container in result.spec.template.spec.containers
                if container.name == 'tomcat'
            ]
            self.assertEqual('tomcat:8', found[0])

            # Scale deployment
            spec = {
                'replicas': 3
            }
            scale_deployment(deploy_name, spec)
            result = get_deployment(deploy_name)
            self.assertEqual(deploy_name, result.metadata.name)
            self.assertEqual(3, result.spec.replicas)

        finally:

            delete_deployment(deploy_name)
            delete_secret(secret_name)
            delete_namespace()

    def test_delete_bogus_deployment(self):
        set_default_namespace('default')
        self.assertEqual(None, delete_deployment('bogus'))




from unittest import TestCase
from k9.helm import *
from k9.core import set_default_namespace, delete_namespace

from k9.pretty_object import PrettyObject

po = PrettyObject()


class TestHelm(TestCase):

    def test_helm_repo_add(self):
        set_default_namespace('unittest')
        result = helm_repo_add('simoncomputing', 'http://charts.simoncomputing.com')
        self.assertTrue("has been added" in result.stdout)

        helm_repo_update()

        result = helm_repo_ls()

        found = [
            repo['name']
            for repo in result
            if repo['name'] == 'simoncomputing'
        ]

        self.assertEqual(1, len(found))

        result = helm_repo_remove('simoncomputing')
        self.assertTrue("has been removed" in result.stdout)

    def test_helm_repo_add_fail(self):
        with self.assertRaisesRegex(Exception, "non-zero exit status 1."):
            helm_repo_add('simoncomputing', 'http://bogus')

    def test_helm_repo_remove_fail(self):
        with self.assertRaisesRegex(Exception, "non-zero exit status 1."):
            helm_repo_remove('bogus')

    def test_helm_install(self):
        try:
            set_default_namespace('tomcat')

            helm_repo_add('stable', 'https://kubernetes-charts.storage.googleapis.com/')

            # Test helm_install()
            helm_install('stable/tomcat', {'domain': 'sandbox.simoncomputing.com'})

            # test helm_ls()
            release_name = 'tomcat'
            result = helm_ls()
            found = [
                release
                for release in result
                if release['name'] == release_name
            ]
            self.assertIsNotNone(found)
            self.assertEqual(release_name, found[0]['name'])

            # test helm_exists()
            self.assertTrue(helm_exists(release_name))

        finally:
            # test helm_uninstall()
            result = helm_uninstall(release_name)
            print(f'result={result.stdout}')
            self.assertTrue('uninstalled' in result.stdout)
            self.assertFalse(helm_exists(release_name))

            helm_repo_remove('stable')
            delete_namespace()

    def test_helm_uninstall_fail(self):
        with self.assertRaisesRegex(Exception, 'returned non-zero exit status'):
            helm_uninstall('bogus')






from unittest import TestCase, mock
from k9.core import *
from k9.rbac import *
from k9.batch import *

class TestBatch(TestCase):

    ###########################################################################
    # cron job
    ###########################################################################
    def test_cron_job(self):
        try:
            ###############
            # Arrange
            set_default_namespace('default')
            sa_name = 'xx-ecr-cron-sa'
            bind_name = 'xx-ecr-cron-sa-bind'
            secret_name = 'aws-info'
            cj_name = 'xx-ecr-login-cron'
            role_name = 'ecr-login-role'

            create_service_account(sa_name)
            self.assertTrue(service_account_exists(sa_name))

            role = {
                'apiVersion': 'rbac.authorization.k8s.io/v1',
                'kind': 'ClusterRole',
                'metadata': {'name': f'{role_name}'},
                'rules': [
                    {
                        'apiGroups': [''],
                        'resources': ["serviceaccounts"],
                        'verbs': ["get", "patch"]

                    },
                    {
                        'apiGroups': [''],
                        'resources': ['secrets'],
                        'verbs': ['create', 'delete']
                    }
                 ]
            }
            create_cluster_role(role)
            self.assertTrue(cluster_role_exists(role_name))

            create_cluster_role_binding(bind_name, role_name, sa_name)
            self.assertTrue(cluster_role_binding_exists(bind_name))

            secrets = {
                'ecr-url': 'https://some-url',
                'aws-account': 'bogus account value',
                'aws-region': 'us-east-1'
            }
            create_secret(secret_name, secrets)
            self.assertTrue(secret_exists(secret_name))

            ###############
            # Act
            body = read_yaml(abs_path('../test/ecr-login-cron.yml'))
            result = create_cron_job(body)

            ###############
            # Assert
            self.assertEqual(cj_name, result.metadata.name)
            self.assertTrue(cron_job_exists(cj_name))

            # test list_cron_jobs()
            found = [
                cj['name']
                for cj in list_cron_jobs()
                if cj['name'] == cj_name
            ]
            self.assertEqual(1, len(found))

            # test get_cron_job()
            result = get_cron_job(cj_name)
            self.assertEqual(cj_name, result.metadata.name)

        finally:
            delete_cron_job(cj_name)
            delete_secret(secret_name)
            delete_cluster_role_binding(bind_name)
            delete_cluster_role(role_name)
            delete_service_account(sa_name)


    def test_cron_job_exists_fail(self):
        self.assertFalse(cron_job_exists('bogus'))

    def test_delete_bogus_cron_job(self):
        set_default_namespace('default')
        self.assertEqual(None, delete_cron_job('bogus'))
